//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"encoding/json"
	"io/ioutil"
	"log"
	"net/http"
	"strings"

	"github.com/gin-gonic/gin"
)

var auth = &Auth{}

type Auth struct {
	sessions *SessionManager
	oidc     *OIDCBackend
	infoLog  *log.Logger
	errLog   *log.Logger
	dbgLog   *log.Logger
}

func Init(c *Config, infoLog, errLog, dbgLog *log.Logger) (err error) {
	if infoLog == nil {
		infoLog = log.New(ioutil.Discard, "", 0)
	}
	if errLog == nil {
		errLog = log.New(ioutil.Discard, "", 0)
	}
	if dbgLog == nil {
		dbgLog = log.New(ioutil.Discard, "", 0)
	}

	if c == nil {
		infoLog.Println("authentication: disabled - will allow any request!")
		return
	}
	auth.infoLog = infoLog
	auth.errLog = errLog
	auth.dbgLog = dbgLog

	if auth.sessions, err = NewSessionManager(c.Sessions); err != nil {
		auth.errLog.Printf("authentication: failed to initialize session manager: %v", err)
		return
	}

	if c.OIDC != nil {
		if auth.oidc, err = NewOIDCBackend(c.OIDC); err != nil {
			auth.errLog.Printf("authentication: failed to initialize OIDC backend: %v", err)
			return
		}
		auth.infoLog.Printf("authentication: successfully initialized %s", auth.oidc)
	}

	auth.infoLog.Println("authentication: initialized")
	return
}

func newSession(c *gin.Context) {
	request := &NewSessionRequest{}
	err := json.NewDecoder(c.Request.Body).Decode(request)
	if err != nil {
		c.JSON(http.StatusBadRequest, HTTPErrorResponse{"Error decoding request: " + err.Error()})
		return
	}

	response := &NewSessionResponse{}
	switch strings.ToLower(request.Backend) {
	case "oidc":
		if auth.oidc == nil {
			c.JSON(http.StatusConflict, HTTPErrorResponse{"OIDC authentication is not enabled"})
			return
		}
		s, err := auth.oidc.NewOIDCSession(c.Request.Context(), request.Arguments)
		if err != nil {
			c.JSON(http.StatusBadRequest, HTTPErrorResponse{"Error creating session: " + err.Error()})
			return
		}
		response.Session = s
		response.Token = s.getToken()
	default:
		c.JSON(http.StatusBadRequest, HTTPErrorResponse{"invalid authentication backend: " + request.Backend})
		return
	}
	c.JSON(http.StatusOK, response)
}

func getSession(c *gin.Context) {
	s := getSessionFromBearerToken(c.Request)
	if s == nil {
		sendHTTPInvalidSessionResponse(c)
		return
	}

	if c.Query("wait-for-login") == "" {
		c.JSON(http.StatusOK, s)
		return
	}

	ctx := c.Request.Context()
	for {
		s, sub := auth.sessions.getAndSubscribe(s.ID())
		if s == nil {
			c.JSON(http.StatusNotFound, HTTPErrorResponse{"this session does no longer exist"})
			return
		}
		st := s.State()
		if st >= SessionStateLoggedIn {
			c.JSON(http.StatusOK, s)
			return
		}
		select {
		case <-sub:
		case <-ctx.Done():
			c.JSON(http.StatusRequestTimeout, HTTPErrorResponse{"error while waiting for state change: " + ctx.Err().Error()})
			return
		}
	}
}

func deleteSession(c *gin.Context) {
	s := getSessionFromBearerToken(c.Request)
	if s == nil {
		sendHTTPInvalidSessionResponse(c)
		return
	}
	s.setState(SessionStateLoggedOut)
	c.JSON(http.StatusOK, "you are now logged out")
}

func listBackends(c *gin.Context) {
	backends := []AuthBackendInfo{}
	if auth.oidc != nil {
		backend := AuthBackendInfo{Name: "oidc"}
		backend.Description = auth.oidc.String()
		backends = append(backends, backend)
	}

	c.JSON(http.StatusOK, backends)
}

func disabled(c *gin.Context) {
	code := http.StatusBadRequest
	if c.Param("path") == "/backends" {
		code = http.StatusNotFound
	}
	c.JSON(code, HTTPErrorResponse{"authentication is disabled"})
}

func InstallHTTPHandler(r *gin.RouterGroup) {
	if auth.sessions == nil {
		r.Any("/*path", disabled)
		return
	}

	backends := r.Group("backends")
	backends.GET("", listBackends)

	session := r.Group("session")
	session.POST("", newSession)
	session.GET("", getSession)
	session.DELETE("", deleteSession)

	if auth.oidc != nil {
		oidc := r.Group("oidc")
		oidc.GET("login", auth.oidc.Login)
		oidc.GET("callback", auth.oidc.Callback)
	}
}

func Middleware() gin.HandlerFunc {
	if auth.sessions == nil {
		return func(c *gin.Context) {
			c.Request = attachSessionToRequest(c.Request, anonAllowAll)
		}
	}

	return func(c *gin.Context) {
		s := getSessionFromBearerToken(c.Request)
		if s == nil || s.State() != SessionStateLoggedIn {
			sendHTTPInvalidSessionResponse(c)
			c.Abort()
			return
		}
		c.Request = attachSessionToRequest(c.Request, s)
	}
}
